#!/usr/bin/python
# -*- coding: utf-8 -*-

'''
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
'''

from ..allocator import ConstExpr, Evaluate, IsConstExpr
from eudplib import utils as ut


class Condition(ConstExpr):

    '''
    Condition class.

    Memory layout:

     ======  =============  ========  ===========
     Offset  Field name     Position  EPD Player
     ======  =============  ========  ===========
       +00   locid           dword0   EPD(cond)+0
       +04   player          dword1   EPD(cond)+1
       +08   amount          dword2   EPD(cond)+2
       +0C   unitid          dword3   EPD(cond)+3
       +0E   comparison
       +0F   condtype
       +10   restype         dword4   EPD(cond)+4
       +11   flags
       +12   internal[2]
     ======  =============  ========  ===========
    '''

    def __init__(self, locid, player, amount, unitid,
                 comparison, condtype, restype, flags):
        super().__init__(self)

        self.locid = locid
        self.player = player
        self.amount = amount
        self.unitid = unitid
        self.comparison = comparison
        self.condtype = condtype
        self.restype = restype
        self.flags = flags

        self.parenttrg = None
        self.condindex = None

    def Disable(self):
        self.flags |= 2

    # -------

    def CheckArgs(self, i):
        ut.ep_assert(
            self.locid is None or IsConstExpr(self.locid),
            'Invalid arg %s in trigger index %d' % (self.locid, i)
        )
        ut.ep_assert(
            self.player is None or IsConstExpr(self.player),
            'Invalid arg %s in trigger index %d' % (self.player, i)
        )
        ut.ep_assert(
            self.amount is None or IsConstExpr(self.amount),
            'Invalid arg %s in trigger index %d' % (self.amount, i)
        )
        ut.ep_assert(
            self.unitid is None or IsConstExpr(self.unitid),
            'Invalid arg %s in trigger index %d' % (self.unitid, i)
        )
        ut.ep_assert(
            self.comparison is None or IsConstExpr(self.comparison),
            'Invalid arg %s in trigger index %d' % (self.comparison, i)
        )
        ut.ep_assert(
            self.condtype is None or IsConstExpr(self.condtype),
            'Invalid arg %s in trigger index %d' % (self.condtype, i)
        )
        ut.ep_assert(
            self.restype is None or IsConstExpr(self.restype),
            'Invalid arg %s in trigger index %d' % (self.restype, i)
        )
        ut.ep_assert(
            self.flags is None or IsConstExpr(self.flags),
            'Invalid arg %s in trigger index %d' % (self.flags, i)
        )
        return True

    def SetParentTrigger(self, trg, index):
        ut.ep_assert(
            self.parenttrg is None,
            'Condition cannot be shared by two triggers. '
        )

        ut.ep_assert(trg is not None, 'Trigger should not be null.')
        ut.ep_assert(0 <= index < 16, 'WTF')

        self.parenttrg = trg
        self.condindex = index

    def Evaluate(self):
        ut.ep_assert(self.parenttrg is not None, 'Orphan condition')
        return Evaluate(self.parenttrg) + 8 + self.condindex * 20

    def WritePayload(self, pbuffer):
        pbuffer.WritePack(
            'IIIHBBBBH',
            self.locid,
            self.player,
            self.amount,
            self.unitid,
            self.comparison,
            self.condtype,
            self.restype,
            self.flags,
            0
        )



def StackObjects(
    found_objects,
    dwoccupmap_dict,
    alloctable,
):
    dwoccupmap_max_size = 0
    for obj in found_objects:
        dwoccupmap_max_size += len(dwoccupmap_dict[obj])

    # Buffer to sum all dwoccupmaps
    dwoccupmap_sum = [-1] * (dwoccupmap_max_size + 1)
    dwoccupmap = [0] * (dwoccupmap_max_size + 1)
    lallocaddr = 0
    payload_size = 0

    for obj in found_objects:
        # Convert to faster c array
        py_dwoccupmap = dwoccupmap_dict[obj]
        oclen = len(py_dwoccupmap)
        for j in range(oclen):
            dwoccupmap[j] = py_dwoccupmap[j]

        # preprocess dwoccupmap
        for i in range(oclen):
            if dwoccupmap[i] == 0:
                dwoccupmap[i] = -1
            elif i == 0 or dwoccupmap[i - 1] == -1:
                dwoccupmap[i] = i
            else:
                dwoccupmap[i] = dwoccupmap[i - 1]

        # Find appropriate position to allocate object
        j = 0
        while j < oclen:
            # Update on conflict map
            if dwoccupmap[j] != -1 and dwoccupmap_sum[lallocaddr + j] != -1:
                lallocaddr = dwoccupmap_sum[lallocaddr + j] - dwoccupmap[j]
                j = 0

            else:
                j += 1

        # Apply occupation map
        for j in range(oclen - 1, -1, -1):
            curoff = lallocaddr + j
            if dwoccupmap[j] != -1 or dwoccupmap_sum[curoff] != -1:
                if dwoccupmap_sum[curoff + 1] == -1:
                    dwoccupmap_sum[curoff] = curoff + 1
                else:
                    dwoccupmap_sum[curoff] = dwoccupmap_sum[curoff + 1]

        alloctable[obj] = lallocaddr * 4
        if (lallocaddr + oclen) * 4 > payload_size:
            payload_size = (lallocaddr + oclen) * 4



#!/usr/bin/python
# -*- coding: utf-8 -*-

'''
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
'''

import weakref
import collections

from .. import rawtrigger as bt
from ..allocator import (
    Evaluate,
    Forward,
    ConstExpr,
    IsConstExpr,
)

from ...utils import (
    EPD,
    ExprProxy,
    ep_assert,
    unProxy,
    cachedfunc
)

from ..variable import EUDVariable, SeqCompute
from ..variable.vbuf import GetCurrentVariableBuffer


@cachedfunc
def EUDVArrayData(size):
    ep_assert(isinstance(size, int))

    class _EUDVArrayData(ConstExpr):
        def __init__(self, initvars):
            super().__init__(self)
            ep_assert(len(initvars) == size, "%d items expected" % size)
            self._initvars = initvars
            self._vdict = weakref.WeakKeyDictionary()

        def Evaluate(self):
            evb = GetCurrentVariableBuffer()
            if evb not in self._vdict:
                variables = [
                    evb.CreateVarTrigger(ival) for ival in self._initvars]
                self._vdict[evb] = variables[0]

            return Evaluate(self._vdict[evb])

    return _EUDVArrayData


@cachedfunc
def EUDVArray(size, basetype=None):
    ep_assert(isinstance(size, int))

    class _EUDVArray(ExprProxy):
        def __init__(self, initvars=None):
            # Int -> interpret as sequence of 0s
            if isinstance(initvars, int):
                initvars = [0] * size

            # For python iterables
            if isinstance(initvars, collections.Iterable):
                baseobj = EUDVArrayData(size)(initvars)

            # Initialization by constant reference
            elif IsConstExpr(initvars):
                initvars = unProxy(initvars)
                ep_assert(isinstance(initvars, EUDVArrayData(size)))
                baseobj = initvars

            # Initialization by variable reference
            else:
                baseobj = EUDVariable()
                baseobj << initvars

            super().__init__(baseobj)
            self.dontFlatten = True
            self._epd = EPD(self)
            self._basetype = basetype

        def getItemPtr(self, i):
            return self + 72 * i

        def getItemEPD(self, i):
            return self._epd + 18 * i

        def get(self, i):
            # This function is hand-optimized

            r = EUDVariable()
            itemptr = self.getItemPtr(i)
            itemepd = self.getItemEPD(i)

            vtproc = Forward()
            nptr = Forward()
            a0, a1, a2 = Forward(), Forward(), Forward()

            SeqCompute([
                (EPD(vtproc + 4), bt.SetTo, itemptr),
                (EPD(a0 + 16), bt.SetTo, itemepd + (8 + 320 + 16) // 4),
                (EPD(a1 + 16), bt.SetTo, itemepd + (8 + 320 + 24) // 4),
                (EPD(a2 + 16), bt.SetTo, itemepd + 1),
            ])

            vtproc << bt.RawTrigger(
                nextptr=0,
                actions=[
                    a0 << bt.SetDeaths(0, bt.SetTo, EPD(r.getValueAddr()), 0),
                    a1 << bt.SetDeaths(0, bt.SetTo, 0x072D0000, 0),
                    a2 << bt.SetDeaths(0, bt.SetTo, nptr, 0),
                ]
            )

            nptr << bt.NextTrigger()
            if self._basetype:
                r = self._basetype(r)
            return r

        def set(self, i, value):
            itemepd = self.getItemEPD(i)
            a0, t = Forward(), Forward()
            SeqCompute([
                (EPD(a0 + 16), bt.SetTo, itemepd + (8 + 320 + 20) // 4),
                (EPD(a0 + 20), bt.SetTo, value),
            ])
            t << bt.RawTrigger(
                actions=[
                    a0 << bt.SetDeaths(0, bt.SetTo, 0, 0),
                ]
            )

        def fill(self, values, *, assert_expected_values_len=None):
            if assert_expected_values_len:
                ep_assert(len(values) == assert_expected_values_len)

            SeqCompute([
                (EPD(self + 344 + i * 72), bt.SetTo, value)
                for i, value in enumerate(values)
            ])

        def __getitem__(self, i):
            return self.get(i)

        def __setitem__(self, i, value):
            return self.set(i, value)

    return _EUDVArray



#!/usr/bin/python
# -*- coding: utf-8 -*-

'''
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
'''

import weakref
import traceback

from .. import rawtrigger as bt
from ..allocator import (
    Evaluate,
    Forward,
    ConstExpr,
    IsConstExpr
)
from ...utils import (
    FlattenList,
    EPD,
    List2Assignable,
    unProxy,
    ep_assert,
    EPError
)
from .vbase import VariableBase
from .vbuf import GetCurrentVariableBuffer


# Unused variable don't need to be allocated.
class VariableTriggerForward(ConstExpr):

    def __init__(self, initval):
        super().__init__(self)
        self._vdict = weakref.WeakKeyDictionary()
        self._initval = initval

    def Evaluate(self):
        evb = GetCurrentVariableBuffer()
        if evb not in self._vdict:
            self._vdict[evb] = evb.CreateVarTrigger(self._initval)

        return Evaluate(self._vdict[evb])


class EUDVariable(VariableBase):

    '''
    Full variable.
    '''

    def __init__(self, initval=0):
        self._vartrigger = VariableTriggerForward(initval)
        self._varact = self._vartrigger + (8 + 320)
        self._rvalue = False

    def GetVTable(self):
        return self._vartrigger

    def getValueAddr(self):
        return self._varact + 20

    def __hash__(self):
        return id(self)

    # -------
    def makeL(self):
        self._rvalue = False
        return self

    def makeR(self):
        self._rvalue = True
        return self

    def checkNonRValue(self):
        if self._rvalue:
            raise EPError('Trying to modify value of l-value variable')

    # -------

    def QueueAssignTo(self, dest):
        try:
            dest.checkNonRValue()
            dest = EPD(dest.getValueAddr())
        except AttributeError:
            pass

        return [
            bt.SetDeaths(EPD(self._varact + 16), bt.SetTo, dest, 0),
            bt.SetDeaths(EPD(self._varact + 24), bt.SetTo, 0x072D0000, 0),
        ]

    def QueueAddTo(self, dest):
        try:
            dest.checkNonRValue()
            dest = EPD(dest.getValueAddr())
        except AttributeError:
            pass

        return [
            bt.SetDeaths(EPD(self._varact + 16), bt.SetTo, dest, 0),
            bt.SetDeaths(EPD(self._varact + 24), bt.SetTo, 0x082D0000, 0),
        ]

    def QueueSubtractTo(self, dest):
        try:
            dest.checkNonRValue()
            dest = EPD(dest.getValueAddr())
        except AttributeError:
            pass

        return [
            bt.SetDeaths(EPD(self._varact + 16), bt.SetTo, dest, 0),
            bt.SetDeaths(EPD(self._varact + 24), bt.SetTo, 0x092D0000, 0),
        ]

    # -------

    def Assign(self, other):
        self.checkNonRValue()
        SeqCompute((
            (self, bt.SetTo, other),
        ))

    def __lshift__(self, other):
        self.Assign(other)
        return self

    def __iadd__(self, other):
        SeqCompute(((self, bt.Add, other),))
        return self

    def __isub__(self, other):
        self << self - other
        return self

    # -------

    def __add__(self, other):
        t = EUDVariable()
        SeqCompute([
            (t, bt.SetTo, self),
            (t, bt.Add, other)
        ])
        return t.makeR()

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        t = EUDVariable()

        SeqCompute([
            (t, bt.SetTo, 0xffffffff),
            (t, bt.Subtract, other),
            (t, bt.Add, 1),
            (t, bt.Add, self),
        ])
        return t.makeR()

    def __rsub__(self, other):
        t = EUDVariable()
        SeqCompute([
            (t, bt.SetTo, 0xffffffff),
            (t, bt.Subtract, self),
            (t, bt.Add, 1),
            (t, bt.Add, other),
        ])
        return t.makeR()

    def __neg__(self):
        return (0 - self).makeR()

    def __invert__(self):
        t = EUDVariable()
        SeqCompute([
            (t, bt.SetTo, 0xffffffff),
            (t, bt.Subtract, self)
        ])
        return t.makeR()

    # -------

    def __eq__(self, other):
        if IsConstExpr(other):
            return self.Exactly(other)

        else:
            return (self - other).Exactly(0)

    def __ne__(self, other):
        if other == 0:
            return self.AtLeast(1)
        else:
            return (self - other).AtLeast(1)

    def __le__(self, other):
        if IsConstExpr(other):
            return self.AtMost(other)

        else:
            t = EUDVariable()
            SeqCompute((
                (t, bt.SetTo, self),
                (t, bt.Subtract, other)
            ))
            return t.Exactly(0)

    def __ge__(self, other):
        if IsConstExpr(other):
            return self.AtLeast(other)

        else:
            t = EUDVariable()
            SeqCompute((
                (t, bt.SetTo, other),
                (t, bt.Subtract, self)
            ))
            return t.Exactly(0)

    def __lt__(self, other):
        if isinstance(other, int) and other <= 0:
            print('[Warning] No unsigned number can be leq than %d' % other)
            traceback.print_stack()
            return [bt.Never()]  # No unsigned number is less than 0

        if IsConstExpr(other):
            return self.AtMost(other - 1)

        else:
            t = EUDVariable()
            SeqCompute((
                (t, bt.SetTo, 1),
                (t, bt.Add, self),
                (t, bt.Subtract, other)
            ))
            return t.Exactly(0)

    def __gt__(self, other):
        if isinstance(other, int) and other >= 0xFFFFFFFF:
            print('[Warning] No unsigned int can be greater than %d' % other)
            traceback.print_stack()
            return [bt.Never()]  # No unsigned number is less than 0

        if IsConstExpr(other):
            return self.AtLeast(other + 1)

        else:
            t = EUDVariable()
            SeqCompute((
                (t, bt.SetTo, self),
                (t, bt.Subtract, other)
            ))
            return t.AtLeast(1)

    # operator placeholders
    def __mul__(self, a):
        pass

    def __rmul__(self, a):
        pass

    def __imul__(self, a):
        pass

    def __floordiv__(self, a):
        raise NotImplementedError('')

    def __rfloordiv__(self, a):
        pass

    def __ifloordiv__(self, a):
        pass

    def __mod__(self, a):
        pass

    def __rmod__(self, a):
        pass

    def __imod__(self, a):
        pass

    def __and__(self, a):
        pass

    def __rand__(self, a):
        pass

    def __iand__(self, a):
        pass

    def __or__(self, a):
        pass

    def __ror__(self, a):
        pass

    def __ior__(self, a):
        pass

    def __xor__(self, a):
        pass

    def __rxor__(self, a):
        pass

    def __ixor__(self, a):
        pass


def IsEUDVariable(x):
    x = unProxy(x)
    return isinstance(x, EUDVariable)


# ---------


def VProc(v, actions):
    nexttrg = Forward()

    bt.RawTrigger(
        nextptr=v.GetVTable(),
        actions=[actions] + [bt.SetNextPtr(v.GetVTable(), nexttrg)]
    )

    nexttrg << bt.NextTrigger()


# From vbuffer.py
def EUDCreateVariables(varn):
    return List2Assignable([EUDVariable() for _ in range(varn)])


# -------

def _GetComputeDest(dst):
    try:
        return EPD(dst.getValueAddr())
    except AttributeError:
        return dst


def _SeqComputeSub(assignpairs):
    """
    Subset of SeqCompute with following restrictions

    - Assignment from variable should be after assignment from constant.
    - Total number of actions should be leq than 64
    """

    actionlist = []

    # Collect constant-assigning actions
    const_assigning_index = len(assignpairs)

    for i, assignpair in enumerate(assignpairs):
        dst, mdt, src = assignpair
        if IsEUDVariable(src):
            const_assigning_index = i
            break

    for dst, mdt, src in assignpairs[0:const_assigning_index]:
        dst = _GetComputeDest(dst)
        actionlist.append(bt.SetDeaths(dst, mdt, src, 0))

    # Only constant-assigning function : skip
    if const_assigning_index == len(assignpairs):
        bt.RawTrigger(actions=actionlist)
        return

    #
    # Rest is for non-constant assigning actions
    #
    nextptr = None  # nextptr for this rawtrigger
    vt_nextptr = None  # what to set for nextptr of current vtable

    for dst, mdt, src in assignpairs[const_assigning_index:]:
        dst = _GetComputeDest(dst)

        if nextptr is None:
            nextptr = src.GetVTable()
        else:
            vt_nextptr << src.GetVTable()

        vt_nextptr = Forward()
        queuef = {
            bt.SetTo: EUDVariable.QueueAssignTo,
            bt.Add: EUDVariable.QueueAddTo,
            bt.Subtract: EUDVariable.QueueSubtractTo,
        }[mdt]

        actionlist.append([
            queuef(src, dst),
            bt.SetNextPtr(src.GetVTable(), vt_nextptr)
        ])

    bt.RawTrigger(
        nextptr=nextptr,
        actions=actionlist
    )

    vt_nextptr << bt.NextTrigger()


def SeqCompute(assignpairs):
    # We need dependency map while writing assignment pairs
    dstvarset = set()
    srcvarset = set()

    # Sublist of assignments to put in _SeqComputeSub
    subassignpairs = []

    # Is we collecting constant-assigning pairs?
    constcollecting = True

    # Number of expected actions.
    actioncount = 0

    def FlushPairs():
        nonlocal constcollecting, actioncount

        if actioncount == 0:  # Already flushed before
            return

        _SeqComputeSub(subassignpairs)

        dstvarset.clear()
        srcvarset.clear()
        subassignpairs.clear()
        constcollecting = True
        actioncount = 0

    for assignpair in assignpairs:
        dst, mdt, src = assignpair
        dst = unProxy(dst)
        src = unProxy(src)

        # Flush action set before preceeding
        if IsEUDVariable(src):
            if src in dstvarset:
                FlushPairs()
            elif src in srcvarset:
                FlushPairs()
            elif actioncount >= 64 - 3:
                FlushPairs()

            srcvarset.add(src)
            constcollecting = False
            actioncount += 3

        else:
            if not constcollecting:
                FlushPairs()
            elif actioncount >= 64 - 3:
                FlushPairs()

            actioncount += 1

        subassignpairs.append(assignpair)
        if IsEUDVariable(dst):
            dstvarset.add(dst)

    FlushPairs()


def SetVariables(srclist, dstlist, mdtlist=None):
    srclist = FlattenList(srclist)
    dstlist = FlattenList(dstlist)
    ep_assert(len(srclist) == len(dstlist), 'Input/output size mismatch')

    if mdtlist is None:
        mdtlist = [bt.SetTo] * len(srclist)

    sqa = [(src, mdt, dst) for src, dst, mdt in zip(srclist, dstlist, mdtlist)]
    SeqCompute(sqa)



#!/usr/bin/python
# -*- coding: utf-8 -*-

'''
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
'''

''' Stage 1 : works in TRIG section.
- Check if EUD is available. If not, defeat
- Initialize stage 2 & execute it
- Modifies TRIG rawtrigger's nextptr. Modification is fixed in stage 3.
'''

from ... import core as c
from eudplib import utils as ut
from ...trigtrg import trigtrg as tt


trglist = []


def Trigger(*args, **kwargs):
    global trglist
    trglist.append(tt.Trigger(*args, **kwargs))


def CopyDeaths(iplayer, oplayer, copyepd=False, initvalue=None):
    tmpplayer = 11

    if initvalue is None:
        if copyepd:
            initvalue = tt.EPD(0)
        else:
            initvalue = 0

    Trigger(
        actions=[
            tt.SetDeaths(oplayer, tt.SetTo, initvalue, 0),
            tt.SetDeaths(tmpplayer, tt.SetTo, 0, 0),
        ]
    )

    for i in range(31, 1, -1):
        if copyepd:
            subval = 2 ** (i - 2)

        else:
            subval = 2 ** i

        Trigger(
            conditions=[tt.Deaths(iplayer, tt.AtLeast, 2 ** i, 0)],
            actions=[
                tt.SetDeaths(iplayer, tt.Subtract, 2 ** i, 0),
                tt.SetDeaths(tmpplayer, tt.Add, 2 ** i, 0),
                tt.SetDeaths(oplayer, tt.Add, subval, 0),
            ]
        )

    for i in range(31, 1, -1):
        Trigger(
            players=[tt.AllPlayers],
            conditions=[tt.Deaths(tmpplayer, tt.AtLeast, 2 ** i, 0)],
            actions=[
                tt.SetDeaths(iplayer, tt.Add, 2 ** i, 0),
                tt.SetDeaths(tmpplayer, tt.Subtract, 2 ** i, 0),
            ]
        )


def CreateVectorRelocator(chkt, payload):
    global trglist

    # Append 'Require EUD enabler' to string table
    str_section = chkt.getsection('STR')
    strtb = c.TBL(str_section)
    eude_needed = strtb.GetStringIndex('This map requires EUD Enabler to run')
    str_section = strtb.SaveTBL()

    '''
    Algorithm credit to klassical_31@naver.com

    Overall algorithm : STR <-> MRGN cycle

    MRGN:
      SetMemory(payload, Add, payload_offset // 4)

    MRGN <-> STR0  =  Trigger(actions=[
             STR1        [SetMemory(
                            mrgn.action[i].player,
                            Add,
                            prttable[j+packn] - prttable[j]
                          ) for i in range(packn)],
             STR2        SetNextPtr(mrgn, STR[k+1])
             ...      )
    '''

    mrgn = 0x58DC60
    packn = 15

    #############
    # STR SECTION
    #############
    str_sled = []
    sledheader_prt = b'\0\0\0\0' + ut.i2b4(mrgn)
    sledheader_ort = b'\0\0\0\0' + ut.i2b4(mrgn + 2408)

    # apply prt
    prttrg_start = 2408 * len(str_sled)  # = 0
    # mrgn.action[i].player = EPD(payload_offset) + prevoffset[i]
    prevoffset = [-1] * packn
    for i in range(0, len(payload.prttable), packn):
        prts = list(map(lambda x: x // 4, payload.prttable[i: i + packn]))
        prts = prts + [-1] * (packn - len(prts))  # -1 : dummy space
        pch = [prts[j] - prevoffset[j] for j in range(packn)]
        str_sled.append(sledheader_prt + tt.Trigger(
            actions=[
                [tt.SetMemory(mrgn + 328 + 32 * j + 16, tt.Add, pch[j])
                 for j in range(packn)],
            ]
        ))

        prevoffset = prts

    # apply ort
    orttrg_start = 2408 * len(str_sled)  # = 0
    # mrgn.action[i].player = EPD(payload_offset) + prevoffset[i]
    prevoffset = [-1] * packn
    for i in range(0, len(payload.orttable), packn):
        orts = list(map(lambda x: x // 4, payload.orttable[i: i + packn]))
        orts = orts + [-1] * (packn - len(orts))  # -1 : dummy space
        pch = [orts[j] - prevoffset[j] for j in range(packn)]
        str_sled.append(sledheader_ort + tt.Trigger(
            actions=[
                [tt.SetMemory(mrgn + 2408 + 328 + 32 * j + 16, tt.Add, pch[j])
                 for j in range(packn)],
            ]
        ))

        prevoffset = orts

    # jump to ort
    str_sled.append(sledheader_ort + tt.Trigger(
        actions=[
            [tt.SetMemory(mrgn + 2408 + 328 + 32 * j + 16, tt.Add,
                          0xFFFFFFFF - prevoffset[j]) for j in range(packn)],
            tt.SetMemory(mrgn + 2408 + 4, tt.Add, 4)  # skip garbage area
        ]
    ))

    # sled completed
    str_sled = b''.join(str_sled)

    str_padding_length = -len(str_section) & 3
    strsled_offset = len(str_section) + str_padding_length
    payload_offset = strsled_offset + len(str_sled) + 4
    str_section = str_section + \
        bytes(str_padding_length) + str_sled + b'\0\0\0\0' + payload.data
    chkt.setsection('STR', str_section)

    ##############
    # MRGN SECTION
    ##############
    mrgn_trigger = []
    mrgn_trigger.append(
        bytes(4) + ut.i2b4(prttrg_start + strsled_offset) +
        tt.Trigger(
            actions=[
                # SetDeaths actions in MRGN initially points to EPD(payload -
                # 4)
                [tt.SetMemory(payload_offset - 4, tt.Add, payload_offset // 4)
                 for _ in range(packn)],
                tt.SetMemory(mrgn + 4, tt.Add, 2408)
            ]
        )
    )

    mrgn_trigger.append(
        bytes(4) + ut.i2b4(orttrg_start + strsled_offset) +
        tt.Trigger(
            actions=[
                [tt.SetMemory(payload_offset - 4, tt.Add, payload_offset)
                 for _ in range(packn)],
                tt.SetMemory(mrgn + 2408 + 4, tt.Add, 2408)
            ]
        )
    )

    mrgn_section = b''.join(mrgn_trigger) + bytes(5100 - 2408 * 2)
    chkt.setsection('MRGN', mrgn_section)

    ##############
    # TRIG SECTION
    ##############
    trglist.clear()

    # Check if epd is supported
    Trigger(actions=[
        tt.SetDeaths(-12, tt.SetTo, 1, 1)
    ])

    Trigger(
        conditions=[
            tt.Deaths(0, tt.Exactly, 0, 0)
        ],
        actions=[
            tt.DisplayTextMessage(eude_needed),
            tt.Draw()
        ]
    )

    # End trigger execution if EUDA is not supported
    #  - using condition #115 twice.
    Trigger(
        conditions=[
            tt.Deaths(0, tt.Exactly, 0, 0),
            tt.Condition(0, 0, 0, 0, 0, 115, 0, 0),
            tt.Condition(0, 0, 0, 0, 0, 115, 0, 0),
        ]
    )

    Trigger(actions=[
        tt.SetDeaths(-12, tt.SetTo, 0, 1)
    ])

    # -------

    # Init mrgn rawtrigger
    strs = 0x5993D4
    for e in range(31, 1, -1):
        # prt table
        # player
        Trigger(
            conditions=tt.Memory(strs, tt.AtLeast, 2 ** e),
            actions=[
                [tt.SetMemory(mrgn + 328 + 32 * i + 16, tt.Add, 2 ** (e - 2))
                 for i in range(packn)],
                tt.SetDeaths(11, tt.Add, 2 ** e, 0),
                [tt.SetMemory(mrgn + 328 + 32 * i + 20, tt.Add, 2 ** (e - 2))
                 for i in range(packn)],
                tt.SetMemory(mrgn + 4, tt.Add, 2 ** e),
                [tt.SetMemory(
                    mrgn + 2408 + 328 + 32 * i + 16, tt.Add, 2 ** (e - 2))
                 for i in range(packn)],
                tt.SetMemory(mrgn + 2408 + 4, tt.Add, 2 ** e),
                [tt.SetMemory(mrgn + 2408 + 328 + 32 * i + 20, tt.Add, 2 ** e)
                 for i in range(packn)],
                tt.SetMemory(strs, tt.Subtract, 2 ** e),
            ]
        )

    for e in range(31, 1, -1):
        Trigger(
            conditions=tt.Deaths(11, tt.AtLeast, 2 ** e, 0),
            actions=[
                tt.SetDeaths(11, tt.Subtract, 2 ** e, 0),
                tt.SetMemory(strs, tt.Add, 2 ** e)
            ]
        )

    # Payload update
    curpl = 0x6509B0

    # -------

    # pts[player].lasttrigger->next = value(strs) + strsled_offset

    pts = 0x51A280

    for player in range(8):
        triggerend = ~(0x51A284 + player * 12)

        Trigger(
            players=[player],
            actions=[
                tt.SetMemory(curpl, tt.SetTo, ut.EPD(pts + 12 * player + 4)),
                tt.SetDeaths(9, tt.SetTo, triggerend, 0)  # Used in stage 2
            ]
        )

    # read pts[player].lasttrigger
    for e in range(31, 1, -1):
        Trigger(
            conditions=tt.Deaths(tt.CurrentPlayer, tt.AtLeast, 2 ** e, 0),
            actions=[
                tt.SetDeaths(tt.CurrentPlayer, tt.Subtract, 2 ** e, 0),
                tt.SetDeaths(10, tt.Add, 2 ** e, 0),
                tt.SetDeaths(11, tt.Add, 2 ** e, 0),
            ]
        )

    for e in range(31, 1, -1):
        Trigger(
            conditions=tt.Deaths(10, tt.AtLeast, 2 ** e, 0),
            actions=[
                tt.SetDeaths(10, tt.Subtract, 2 ** e, 0),
                tt.SetDeaths(tt.CurrentPlayer, tt.Add, 2 ** e, 0)
            ]
        )

    # apply to curpl
    Trigger(actions=[
        tt.SetDeaths(10, tt.SetTo, ut.EPD(4), 0),
        tt.SetMemory(curpl, tt.SetTo, ut.EPD(4))
    ])
    for e in range(31, 1, -1):
        Trigger(
            conditions=tt.Deaths(11, tt.AtLeast, 2 ** e, 0),
            actions=[
                tt.SetDeaths(11, tt.Subtract, 2 ** e, 0),
                # used for nextptr recovery in stage 3
                tt.SetDeaths(10, tt.Add, 2 ** (e - 2), 0),
                tt.SetMemory(curpl, tt.Add, 2 ** (e - 2))
            ]
        )

    # now curpl = EPD(value(ptsprev) + 4)
    # value(EPD(value(ptsprev) + 4)) = strs + payload_offset
    CopyDeaths(tt.EPD(strs), tt.CurrentPlayer, False, strsled_offset)

    # Done!
    trigdata = b''.join(trglist)

    # Stage 1 created

    # -------

    # Previous rawtrigger datas

    oldtrigraw = chkt.getsection('TRIG')
    oldtrigs = [oldtrigraw[i:i + 2400]
                for i in range(0, len(oldtrigraw), 2400)]
    proc_trigs = []

    # Collect only enabled triggers
    for trig in oldtrigs:
        trig = bytearray(trig)
        flag = ut.b2i4(trig, 320 + 2048)
        if flag & 8:  # Trigger already disabled
            pass

        flag ^= 8  # Disable it temporarilly. It will be re-enabled at stage 3
        trig[320 + 2048: 320 + 2048 + 4] = ut.i2b4(flag)
        proc_trigs.append(bytes(trig))

    oldtrigraw_processed = b''.join(proc_trigs)
    chkt.setsection('TRIG', trigdata + oldtrigraw_processed)



#!/usr/bin/python
# -*- coding: utf-8 -*-

'''
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
'''

from eudplib import (
    core as c,
    ctrlstru as cs,
    utils as ut,
)

from .rwcommon import br1, bw1
from .dbstr import DBString


@c.EUDFunc
def f_dbstr_addstr(dst, src):
    """Print string as string to dst. Same as strcpy except of return value.

    :param dst: Destination address (Not EPD player)
    :param src: Source address (Not EPD player)

    :returns: dst + strlen(src)
    """
    b = c.EUDVariable()

    br1.seekoffset(src)
    bw1.seekoffset(dst)

    if cs.EUDInfLoop()():
        c.SetVariables(b, br1.readbyte())
        bw1.writebyte(b)
        cs.EUDBreakIf(b == 0)
        dst += 1
    cs.EUDEndInfLoop()

    bw1.flushdword()

    return dst


@c.EUDFunc
def f_dbstr_adddw(dst, number):
    """Print number as string to dst.

    :param dst: Destination address (Not EPD player)
    :param number: DWORD to print

    :returns: dst + strlen(itoa(number))
    """
    bw1.seekoffset(dst)

    skipper = [c.Forward() for _ in range(9)]
    ch = [0] * 10

    # Get digits
    for i in range(10):
        number, ch[i] = c.f_div(number, 10)
        if i != 9:
            cs.EUDJumpIf(number == 0, skipper[i])

    # print digits
    for i in range(9, -1, -1):
        if i != 9:
            skipper[i] << c.NextTrigger()
        bw1.writebyte(ch[i] + b'0'[0])
        dst += 1

    bw1.writebyte(0)  # EOS
    bw1.flushdword()

    return dst


@c.EUDFunc
def f_dbstr_addptr(dst, number):
    """Print number as string to dst.

    :param dst: Destination address (Not EPD player)
    :param number: DWORD to print

    :returns: dst + strlen(itoa(number))
    """
    bw1.seekoffset(dst)
    ch = [0] * 8

    # Get digits
    for i in range(8):
        number, ch[i] = c.f_div(number, 16)

    # print digits
    for i in range(7, -1, -1):
        if cs.EUDIf()(ch[i] <= 9):
            bw1.writebyte(ch[i] + b'0'[0])
        if cs.EUDElse()():
            bw1.writebyte(ch[i] + (b'A'[0] - 10))
        cs.EUDEndIf()
        dst += 1

    bw1.writebyte(0)  # EOS
    bw1.flushdword()

    return dst


class hptr:
    def __init__(self, value):
        self._value = value


def f_dbstr_print(dst, *args):
    """Print multiple string / number to dst.

    :param dst: Destination address (Not EPD player)
    :param args: Things to print

    """
    if isinstance(dst, DBString):
        dst = dst.GetStringMemoryAddr()

    args = ut.FlattenList(args)
    for arg in args:
        arg = ut.unProxy(arg)
        if isinstance(arg, bytes):
            dst = f_dbstr_addstr(dst, c.Db(arg + b'\0'))
        elif isinstance(arg, str):
            dst = f_dbstr_addstr(dst, c.Db(ut.u2b(arg) + b'\0'))
        elif isinstance(arg, DBString):
            dst = f_dbstr_addstr(dst, arg.GetStringMemoryAddr())
        elif isinstance(arg, int):
            # int and c.EUDVariable should act the same if possible.
            # EUDVariable has a value of 32bit unsigned integer.
            # So we adjust arg to be in the same range.
            dst = f_dbstr_addstr(dst, c.Db(
                ut.u2b(str(arg & 0xFFFFFFFF)) + b'\0'))
        elif isinstance(arg, c.EUDVariable):
            dst = f_dbstr_adddw(dst, arg)
        elif c.IsConstExpr(arg):
            dst = f_dbstr_adddw(dst, arg)
        elif isinstance(arg, hptr):
            dst = f_dbstr_addptr(dst, arg._value)
        else:
            raise ut.EPError(
                'Object wit unknown parameter type %s given to f_eudprint.'
                % type(arg)
            )

    return dst


_printf_buffer = DBString(8192)


def f_simpleprint(*args, spaced=True):
    # Add spaces between arguments
    if spaced:
        spaced_args = []
        for arg in args:
            spaced_args.extend([arg, ' '])
        args = spaced_args[:-1]

    # Print
    f_dbstr_print(_printf_buffer, *args)
    cs.DoActions(_printf_buffer.GetDisplayAction())



#!/usr/bin/python
# -*- coding: utf-8 -*-

'''
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
'''

from ..allocator import ConstExpr, Evaluate, IsConstExpr
from eudplib import utils as ut


class Action(ConstExpr):

    '''
    Action class.

    Memory layout.

     ======  ============= ========  ==========
     Offset  Field Name    Position  EPD Player
     ======  ============= ========  ==========
       +00   locid1         dword0   EPD(act)+0
       +04   strid          dword1   EPD(act)+1
       +08   wavid          dword2   EPD(act)+2
       +0C   time           dword3   EPD(act)+3
       +10   player1        dword4   EPD(act)+4
       +14   player2        dword5   EPD(act)+5
       +18   unitid         dword6   EPD(act)+6
       +1A   acttype
       +1B   amount
       +1C   flags          dword7   EPD(act)+7
       +1D   internal[3
     ======  ============= ========  ==========
    '''

    def __init__(self, locid1, strid, wavid, time, player1, player2,
                 unitid, acttype, amount, flags):
        '''
        See :mod:`eudplib.base.stocktrg` for stock actions list.
        '''
        super().__init__(self)

        self.locid1 = locid1
        self.strid = strid
        self.wavid = wavid
        self.time = time
        self.player1 = player1
        self.player2 = player2
        self.unitid = unitid
        self.acttype = acttype
        self.amount = amount
        self.flags = flags

        self.parenttrg = None
        self.actindex = None

    def Disable(self):
        self.flags |= 2

    # -------

    def CheckArgs(self, i):
        ut.ep_assert(
            self.locid1 is None or IsConstExpr(self.locid1),
            'Invalid arg %s in trigger index %d' % (self.locid1, i)
        )
        ut.ep_assert(
            self.strid is None or IsConstExpr(self.strid),
            'Invalid arg %s in trigger index %d' % (self.strid, i)
        )
        ut.ep_assert(
            self.wavid is None or IsConstExpr(self.wavid),
            'Invalid arg %s in trigger index %d' % (self.wavid, i)
        )
        ut.ep_assert(
            self.time is None or IsConstExpr(self.time),
            'Invalid arg %s in trigger index %d' % (self.time, i)
        )
        ut.ep_assert(
            self.player1 is None or IsConstExpr(self.player1),
            'Invalid arg %s in trigger index %d' % (self.player1, i)
        )
        ut.ep_assert(
            self.player2 is None or IsConstExpr(self.player2),
            'Invalid arg %s in trigger index %d' % (self.player2, i)
        )
        ut.ep_assert(
            self.unitid is None or IsConstExpr(self.unitid),
            'Invalid arg %s in trigger index %d' % (self.unitid, i)
        )
        ut.ep_assert(
            self.acttype is None or IsConstExpr(self.acttype),
            'Invalid arg %s in trigger index %d' % (self.acttype, i)
        )
        ut.ep_assert(
            self.amount is None or IsConstExpr(self.amount),
            'Invalid arg %s in trigger index %d' % (self.amount, i)
        )
        ut.ep_assert(
            self.flags is None or IsConstExpr(self.flags),
            'Invalid arg %s in trigger index %d' % (self.flags, i)
        )
        return True

    def SetParentTrigger(self, trg, index):
        ut.ep_assert(
            self.parenttrg is None,
            'Actions cannot be shared by two triggers.'
        )

        ut.ep_assert(trg is not None, 'Trigger should not be null.')
        ut.ep_assert(0 <= index < 64, 'Triggers out of range')

        self.parenttrg = trg
        self.actindex = index

    def Evaluate(self):
        return Evaluate(self.parenttrg) + 8 + 320 + 32 * self.actindex

    def WritePayload(self, pbuffer):
        pbuffer.WritePack(
            'IIIIIIHBBBBH',
            self.locid1,
            self.strid,
            self.wavid,
            self.time,
            self.player1,
            self.player2,
            self.unitid,
            self.acttype,
            self.amount,
            self.flags,
            0,
            0
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-

'''
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
'''

'''
String table manager. Internally used in eudplib.
'''

from eudplib import utils as ut


class TBL:
    def __init__(self, content=None):
        #
        # datatb : table of strings                       : string data table
        # dataindextb : string id -> data id              : string offset table
        # stringmap : string -> representative string id
        #

        self._datatb = []
        self._stringmap = {}
        self._dataindextb = []  # String starts from #1
        self._capacity = 2  # Size of STR section

        if content is not None:
            self.LoadTBL(content)

    def LoadTBL(self, content):
        self._datatb.clear()
        self._stringmap.clear()
        self._capacity = 2

        stringcount = ut.b2i2(content, 0)
        for i in range(stringcount):
            i += 1
            stringoffset = ut.b2i2(content, i * 2)
            send = stringoffset
            while content[send] != 0:
                send += 1

            string = content[stringoffset:send]
            self.AddString(string)

    def AddString(self, string):
        string = ut.u2b(string)  # Starcraft uses multibyte encoding.
        if not isinstance(string, bytes):
            raise ut.EPError('Invalid type for string')

        stringindex = len(self._dataindextb)

        # If duplicate text exist -> just proxy it
        try:
            repr_stringid = self._stringmap[string]
            dataindex = self._dataindextb[repr_stringid]
            self._dataindextb.append(dataindex)
            self._capacity += 2  # just string offset

        # Else -> Create new entry
        except KeyError:
            dataindex = len(self._datatb)
            self._stringmap[string] = stringindex
            self._datatb.append(string)
            self._dataindextb.append(dataindex)
            # string + b'\0' + string offset
            self._capacity += len(string) + 1 + 2

        ut.ep_assert(self._capacity < 65536, 'String table overflow')

        return stringindex

    def GetString(self, index):
        if index == 0:
            return None
        else:
            try:
                dataindex = self._dataindextb[index - 1]
                return self._datatb[dataindex]
            except IndexError:
                return None

    def GetStringIndex(self, string):
        string = ut.u2b(string)
        if not isinstance(string, bytes):
            raise ut.EPError('Invalid type for string')

        try:
            return self._stringmap[string] + 1

        except KeyError:
            return self.AddString(string) + 1

    def SaveTBL(self):
        outbytes = []

        # calculate offset of each string
        stroffset = []
        outindex = 2 * len(self._dataindextb) + 2
        for s in self._datatb:
            stroffset.append(outindex)
            outindex += len(s) + 1

        # String count
        outbytes.append(ut.i2b2(len(self._dataindextb)))

        # String offsets
        for dataidx in self._dataindextb:
            outbytes.append(ut.i2b2(stroffset[dataidx]))

        # String data
        for s in self._datatb:
            outbytes.append(s)
            outbytes.append(b'\0')

        return b''.join(outbytes)



from importlib.machinery import (
    FileFinder,
    SourceFileLoader,
)

from .epscompile import epsCompile
from eudplib.utils import EPError
import sys
import re
from bisect import bisect_right

import types


lineno_regex = re.compile(b' *# \\(Line (\\d+)\\) .+')


def modifyCodeLineno(codeobj, codeMap):
    co_lnotab = codeobj.co_lnotab
    co_firstlineno = codeobj.co_firstlineno

    # Reconstruct code data
    new_lnotab = []
    currentLine = co_firstlineno
    currentMappedLine = codeMap(currentLine)
    for i in range(0, len(co_lnotab), 2):
        bytecodeLen, lineAdvance = co_lnotab[i: i + 2]
        nextLine = currentLine + lineAdvance
        nextMappedLine = codeMap(nextLine)
        newLineAdvance = nextMappedLine - currentMappedLine
        while newLineAdvance >= 0xFF:
            new_lnotab.append(bytes([0, 0xFF]))
            newLineAdvance -= 0xFF
        new_lnotab.append(bytes([bytecodeLen, newLineAdvance]))
        currentLine = nextLine
        currentMappedLine = nextMappedLine

    # For code objects
    new_co_consts = []
    for c in codeobj.co_consts:
        if isinstance(c, types.CodeType):
            c = modifyCodeLineno(c, codeMap)
        new_co_consts.append(c)

    codeobj = types.CodeType(
        codeobj.co_argcount,
        codeobj.co_kwonlyargcount,
        codeobj.co_nlocals,
        codeobj.co_stacksize,
        codeobj.co_flags,
        codeobj.co_code,
        tuple(new_co_consts),
        codeobj.co_names,
        codeobj.co_varnames,
        codeobj.co_filename,
        codeobj.co_name,
        codeMap(co_firstlineno),  # codeobj.co_firstlineno,
        b''.join(new_lnotab),  # codeobj.co_lnotab,
        codeobj.co_freevars,
        codeobj.co_cellvars
    )

    return codeobj


class EPSLoader(SourceFileLoader):
    def get_data(self, path):
        """Return the data from path as raw bytes."""
        fileData = open(path, 'rb').read()
        if path.endswith('.pyc') or path.endswith('.pyo'):
            return fileData
        compiled = epsCompile(path, fileData)
        if compiled is None:
            raise EPError('epScript compiled failed for %s' % path)
        return compiled

    def source_to_code(self, data, path, *, _optimize=-1):
        codeobj = super().source_to_code(data, path, _optimize=_optimize)

        # Read lines from code data
        codeLine = [0]
        codeMap = [0]
        data = data.replace(b'\r\n', b'\n')
        for lineno, line in enumerate(data.split(b'\n')):
            match = lineno_regex.match(line)
            if match:
                codeLine.append(lineno + 1)
                codeMap.append(int(match.group(1)))

        # Reconstruct code data
        def lineMapper(line):
            return codeMap[bisect_right(codeLine, line) - 1]
        codeobj = modifyCodeLineno(codeobj, lineMapper)
        return codeobj


class EPSFinder:
    def __init__(self):
        self._finderCache = {}

    def _getFinder(self, path):
        try:
            return self._finderCache[path]
        except KeyError:
            self._finderCache[path] = FileFinder(path, (EPSLoader, ['.eps']))
            return self._finderCache[path]

    def find_spec(self, fullname, path, target=None):
        if path is None:
            path = sys.path
        for pathEntry in path:
            finder = self._getFinder(pathEntry)
            spec = finder.find_spec(fullname)
            if spec is None:
                continue
            return spec


sys.meta_path.append(EPSFinder())



cdef unsigned int cryptTable[0x500]

cdef _InitCryptTable():
    cdef unsigned int seed = 0x00100001
    cdef unsigned int index1 = 0, index2 = 0, i
    cdef unsigned int temp1, temp2

    for index1 in range(0x100):
        index2 = index1
        for _ in range(5):
            seed = (seed * 125 + 3) % 0x2AAAAB
            temp1 = (seed & 0xFFFF) << 0x10

            seed = (seed * 125 + 3) % 0x2AAAAB
            temp2 = seed % 0xFFFF

            cryptTable[index2] = temp1 | temp2
            index2 += 0x100

_InitCryptTable()

# -----

def DecryptDwords(unsigned int blocks[]):
    cdef unsigned int seed = 0xFFFFFFFF

    for dw in dword:
        seed += cryptTable[0x400 + (key & 0xFF)]
        ch = dw ^ (key + seed)

        key = ((~key << 0x15) + 0x11111111) | (key >> 0x0B)
        seed = ch + seed + (seed << 5) + 3;





void DecryptBlock(void *block, long length, unsigned long key)
{
unsigned long seed = 0xEEEEEEEE, unsigned long ch;
unsigned long *castBlock = (unsigned long *)block;

// Round to longs
length >>= 2;

while(length-- > 0)
{
seed += stormBuffer[0x400 + (key & 0xFF)];
ch = *castBlock ^ (key + seed);

key = ((~key << 0x15) + 0x11111111) | (key >> 0x0B);
seed = ch + seed + (seed << 5) + 3;
*castBlock++ = ch;
}
}


#!/usr/bin/python
# -*- coding: utf-8 -*-

'''
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
'''

from . import tblformat
from ...utils import b2i2, b2i4, u2b, ep_assert


class StringIdMap:
    def __init__(self):
        self._s2id = {}

    def AddItem(self, string, id):
        string = u2b(string)
        if string in self._s2id:  # ambigious string
            self._s2id[string] = None

        else:
            self._s2id[string] = id

    def GetStringIndex(self, string):
        string = u2b(string)
        retid = self._s2id[string]
        ep_assert(retid is not None, "Ambigious string %s" % string)
        return retid


strmap = None
unitmap = None
locmap = None
swmap = None


def IgnoreColor(s):
    return bytes(filter(lambda x: not (0x01 <= x <= 0x1F or x == 0x7F), s))


def InitStringMap(chkt):
    global strmap, unitmap, locmap, swmap

    strmap = tblformat.TBL(chkt.getsection('STR'))
    unitmap = StringIdMap()
    locmap = StringIdMap()
    swmap = StringIdMap()

    unix = chkt.getsection('UNIx')
    mrgn = chkt.getsection('MRGN')
    swnm = chkt.getsection('SWNM')

    # Get location names
    if mrgn:
        locn = len(mrgn) // 20
        for i in range(locn):
            locstrid = b2i2(mrgn, i * 20 + 16)
            locstr = strmap.GetString(locstrid)
            if locstr:
                locmap.AddItem(locstr, i)

    # Get unit names
    if unix:
        for i in range(228):
            unitstrid = b2i2(unix, 3192 + i * 2)
            unitstr = strmap.GetString(unitstrid)
            if unitstr:
                unitmap.AddItem(unitstr, i)
                if unitstr != IgnoreColor(unitstr):
                    unitmap.AddItem(IgnoreColor(unitstr), i)

    # Get switch names
    if swnm:
        for i in range(256):
            switchstrid = b2i4(swnm, i * 4)
            switchstr = strmap.GetString(switchstrid)
            if switchstr:
                swmap.AddItem(switchstr, i)


def GetLocationIndex(l):
    return locmap.GetStringIndex(l)


def GetStringIndex(s):
    return strmap.GetStringIndex(s)


def GetSwitchIndex(s):
    return swmap.GetStringIndex(s)


def GetUnitIndex(u):
    return unitmap.GetStringIndex(u)


def ApplyStringMap(chkt):
    chkt.setsection('STR', strmap.SaveTBL())



#!/usr/bin/python
# -*- coding: utf-8 -*-

'''
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
'''

'''
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
'''


class ExprProxy:

    ''' Class which can contain both constxpr and eudvariable '''

    def __init__(self, initval):
        self._value = initval

    def getValue(self):
        return self._value

    def __hash__(self):
        return id(self)

    # Proxy arithmetic operators
    def __lshift__(self, other):
        return self._value << other

    def __add__(self, other):
        return self._value + other

    def __radd__(self, other):
        return other + self._value

    def __sub__(self, other):
        return self._value - other

    def __rsub__(self, other):
        return other - self._value

    def __mul__(self, k):
        return self._value * k

    def __rmul__(self, k):
        return k * self._value

    def __floordiv__(self, k):
        return self._value // k

    def __rfloordiv__(self, other):
        return k // self._value

    def __and__(self, other):
        return self._value & k

    def __rand__(self, other):
        return other & self._value

    def __or__(self, k):
        return self._value | other

    def __ror__(self, k):
        return other | self._value

    def __xor__(self, other):
        return self._value ^ other

    def __rxor__(self, other):
        return other ^ self._value

    def __neg__(self):
        return -self._value

    def __invert__(self):
        return ~self._value

    def __bool__(self):
        return bool(self._value)

    # Proxy comparison operator

    def __eq__(self, k):
        return self._value == k

    def __ne__(self, k):
        return self._value != k

    def __le__(self, other):
        return self._value <= other

    def __lt__(self, other):
        return self._value < other

    def __ge__(self, other):
        return self._value >= other

    def __gt__(self, other):
        return self._value > other

    # TODO: add inplace operators

    # Proxy other methods
    def __getattr__(self, name):
        return getattr(self._value, name)

    def __getitem__(self, name):
        return self._value[name]

    def __setitem__(self, name, newvalue):
        self._value[name] = newvalue


def unProxy(x):
    try:
        return unProxy(x.getValue())
    except AttributeError:
        return x



#!/usr/bin/python
# -*- coding: utf-8 -*-

'''
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
'''

# This code uses simple LCG algorithm.

from eudplib import core as c
from eudplib import utils as ut
from ..memiof import f_dwbreak, f_dwread_epd

_seed = c.EUDVariable()


def f_getseed():
    t = c.EUDVariable()
    t << _seed
    return t


def f_srand(seed):
    _seed << seed


def f_randomize():
    current_rv = f_dwread_epd(ut.EPD(0x51CA14))
    _seed << current_rv


@c.EUDFunc
def f_rand():
    _seed << c.f_mul(_seed, 1103515245) + 12345
    return f_dwbreak(_seed)[1]  # Only HIWORD is returned


@c.EUDFunc
def f_dwrand():
    seed1 = c.f_mul(_seed, 1103515245) + 12345
    seed2 = c.f_mul(seed1, 1103515245) + 12345
    _seed << seed2

    ret = c.EUDVariable()
    ret << 0

    # HIWORD
    for i in range(31, 15, -1):
        c.RawTrigger(
            conditions=seed1.AtLeast(2 ** i),
            actions=[
                seed1.SubtractNumber(2 ** i),
                ret.AddNumber(2 ** i),
            ]
        )

    # LOWORD
    for i in range(31, 15, -1):
        c.RawTrigger(
            conditions=seed2.AtLeast(2 ** i),
            actions=[
                seed2.SubtractNumber(2 ** i),
                ret.AddNumber(2 ** (i - 16)),
            ]
        )

    return ret



#!/usr/bin/python
# -*- coding: utf-8 -*-

'''
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
'''

from .. import (
    allocator as ac,
    rawtrigger as rt,
    variable as ev,
    eudfunc as ef,
)

from eudplib import utils as ut


def f_mul(a, b):
    """Calculate a * b"""
    if isinstance(a, ev.EUDVariable) and isinstance(b, ev.EUDVariable):
        return _f_mul(a, b)

    elif isinstance(a, ev.EUDVariable):
        return f_constmul(b)(a)

    elif isinstance(b, ev.EUDVariable):
        return f_constmul(a)(b)

    else:
        ret = ev.EUDVariable()
        ret << a * b
        return ret


def f_div(a, b):
    """Calculate (a//b, a%b) """
    if isinstance(b, ev.EUDVariable):
        return _f_div(a, b)

    elif isinstance(a, ev.EUDVariable):
        return f_constdiv(b)(a)

    else:
        if b:
            q, m = a // b, a % b
        else:
            q, m = 0xFFFFFFFF, a
        vq, vm = ev.EUDCreateVariables(2)
        ev.SeqCompute([
            (vq, rt.SetTo, q),
            (vm, rt.SetTo, m)
        ])
        return vq, vm


# -------


def f_constmul(number):
    """
    f_constnum(a)(b) calculates b * a.

    :param number: Constant integer being multiplied to other numbers.
    :return: Function taking one parameter.
    """
    if not hasattr(f_constmul, 'mulfdict'):
        f_constmul.mulfdict = {}

    mulfdict = f_constmul.mulfdict

    try:
        return mulfdict[number]
    except KeyError:
        @ef.EUDFunc
        def _mulf(a):
            ret = ev.EUDVariable()
            ret << 0
            for i in range(31, -1, -1):
                rt.RawTrigger(
                    conditions=a.AtLeast(2 ** i),
                    actions=[
                        a.SubtractNumber(2 ** i),
                        ret.AddNumber(2 ** i * number)
                    ]
                )
            return ret

        mulfdict[number] = _mulf
        return _mulf


def f_constdiv(number):
    """
    f_constdiv(a)(b) calculates (b // a, b % a)

    :param number: Constant integer to divide other numbers by.
    :return: Function taking one parameter.
    """
    if not hasattr(f_constdiv, 'divfdict'):
        f_constdiv.divfdict = {}

    divfdict = f_constdiv.divfdict

    try:
        return divfdict[number]
    except KeyError:
        @ef.EUDFunc
        def _divf(a):
            ret = ev.EUDVariable()
            ret << 0
            for i in range(31, -1, -1):
                # number too big
                if 2 ** i * number >= 2 ** 32:
                    continue

                rt.RawTrigger(
                    conditions=a.AtLeast(2 ** i * number),
                    actions=[
                        a.SubtractNumber(2 ** i * number),
                        ret.AddNumber(2 ** i)
                    ]
                )
            return ret, a

        divfdict[number] = _divf
        return _divf


# -------


@ef.EUDFunc
def _f_mul(a, b):
    ret, y0 = ev.EUDCreateVariables(2)

    # Init
    ev.SeqCompute([
        (ret, rt.SetTo, 0),
        (y0, rt.SetTo, b)
    ])

    chain = [ac.Forward() for _ in range(32)]
    chain_y0 = [ac.Forward() for _ in range(32)]

    # Calculate chain_y0
    for i in range(32):
        ev.SeqCompute((
            (ut.EPD(chain_y0[i]), rt.SetTo, y0),
            (y0, rt.Add, y0)
        ))
        if i <= 30:
            p1, p2, p3 = ac.Forward(), ac.Forward(), ac.Forward()
            p1 << rt.RawTrigger(
                nextptr=p2,
                conditions=a.AtMost(2 ** (i + 1) - 1),
                actions=rt.SetNextPtr(p1, p3)
            )
            p3 << rt.RawTrigger(
                nextptr=chain[i],
                actions=rt.SetNextPtr(p1, p2)
            )
            p2 << rt.NextTrigger()

    # Run multiplication chain
    for i in range(31, -1, -1):
        cy0 = ac.Forward()

        chain[i] << rt.RawTrigger(
            conditions=[
                a.AtLeast(2 ** i)
            ],
            actions=[
                a.SubtractNumber(2 ** i),
                cy0 << ret.AddNumber(0)
            ]
        )

        chain_y0[i] << cy0 + 20

    return ret


@ef.EUDFunc
def _f_div(a, b):
    ret, x = ev.EUDCreateVariables(2)

    # Init
    ev.SeqCompute([
        (ret, rt.SetTo, 0),
        (x, rt.SetTo, b),
    ])

    # Chain ac.Forward decl
    chain_x0 = [ac.Forward() for _ in range(32)]
    chain_x1 = [ac.Forward() for _ in range(32)]
    chain = [ac.Forward() for _ in range(32)]

    # Fill in chain
    for i in range(32):
        ev.SeqCompute([
            (ut.EPD(chain_x0[i]), rt.SetTo, x),
            (ut.EPD(chain_x1[i]), rt.SetTo, x),
        ])

        # Skip if over 0x80000000
        p1, p2, p3 = ac.Forward(), ac.Forward(), ac.Forward()
        p1 << rt.RawTrigger(
            nextptr=p2,
            conditions=x.AtLeast(0x80000000),
            actions=rt.SetNextPtr(p1, p3)
        )
        p3 << rt.RawTrigger(
            nextptr=chain[i],
            actions=rt.SetNextPtr(p1, p2)
        )
        p2 << rt.NextTrigger()

        ev.SeqCompute([
            (x, rt.Add, x),
        ])

    # Run division chain
    for i in range(31, -1, -1):
        cx0, cx1 = ac.Forward(), ac.Forward()
        chain[i] << rt.RawTrigger(
            conditions=[
                cx0 << a.AtLeast(0)
            ],
            actions=[
                cx1 << a.SubtractNumber(0),
                ret.AddNumber(2 ** i)
            ]
        )

        chain_x0[i] << cx0 + 8
        chain_x1[i] << cx1 + 20

    return ret, a  # a : remainder



from ctypes import (
    CDLL,
    c_int, c_char_p,
)

from eudplib.utils import (
    u2b, find_data_file
)

libeps = CDLL(find_data_file('libepScriptLib.dll', __file__))
libeps.compileString.argtypes = [c_char_p, c_char_p]
libeps.freeCompiledResult.argtypes = [c_int]


def epsCompile(modname, bCode):
    modname = u2b(modname)
    output = libeps.compileString(modname, bCode)
    if not output:
        return None
    outputStr = c_char_p(output).value
    libeps.freeCompiledResult(output)
    return outputStr



#!/usr/bin/python
# -*- coding: utf-8 -*-

'''
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
'''

'''
WARNING! This code is speciallized for use on eudplib
- Default player of rawtrigger is 'All Player'.
- Only conditions/actions used in eudplib are declared
- Condition/Action input filtering (& 0xFFFFFFFF thing) are only applied to
  player & number section of Deaths/SetDeaths
Note this when using this code outside of eudplib.
'''

from struct import pack
from eudplib import utils as ut

i2b1, i2b2, i2b4 = ut.i2b1, ut.i2b2, ut.i2b4

# for Deaths
AtLeast = 0
AtMost = 1
Exactly = 10

# for Set Death
SetTo = 7
Add = 8
Subtract = 9

# player enum
CurrentPlayer = 13
AllPlayers = 17

# constructor
_bc_dict = {
    1: i2b1,
    2: i2b2,
    4: i2b4,
    None: (lambda x: x)
}


def FlattenList(l):
    if type(l) is bytes or type(l) is str:
        return [l]

    try:
        ret = []
        for item in l:
            ret.extend(FlattenList(item))
        return ret

    except TypeError:  # l is not iterable
        return [l]


def Condition(locid, player, amount, unitid,
              comparison, condtype, restype, flag):
    if player < 0:
        player += 0x100000000  # EPD

    player &= 0xFFFFFFFF
    amount &= 0xFFFFFFFF

    return pack('<IIIHBBBBBB', locid, player, amount, unitid,
                comparison, condtype, restype, flag, 0, 0)


def Action(locid1, strid, wavid, time, player1,
           player2, unitid, acttype, amount, flags):
    player1 &= 0xFFFFFFFF
    player2 &= 0xFFFFFFFF

    if player1 < 0:
        player1 += 0x100000000  # EPD
    if player2 < 0:
        player2 += 0x100000000  # EPD
    return pack('<IIIIIIHBBBBBB', locid1, strid, wavid, time, player1, player2,
                unitid, acttype, amount, flags, 0, 0, 0)


def Trigger(players=[AllPlayers], conditions=[], actions=[]):
    conditions = FlattenList(conditions)
    actions = FlattenList(actions)

    ut.ep_assert(type(players) is list)
    ut.ep_assert(type(conditions) is list)
    ut.ep_assert(type(actions) is list)
    ut.ep_assert(len(conditions) <= 16)
    ut.ep_assert(len(actions) <= 64)
    peff = bytearray(28)
    for p in players:
        peff[p] = 1

    b = b''.join(
        conditions +
        [bytes(20 * (16 - len(conditions)))] +
        actions +
        [bytes(32 * (64 - len(actions)))] +
        [b'\x04\0\0\0'] +
        [bytes(peff)]
    )
    ut.ep_assert(len(b) == 2400)
    return b


# conditions
def Deaths(player, comparison, number, unit):
    return Condition(
        0x00000000, player, number, unit,
        comparison, 0x0F, 0x00, 0x10)


def Memory(offset, comparison, number):
    ut.ep_assert(offset % 4 == 0)  # only this kind of comparison is possible
    player = EPD(offset)

    if 0 <= player < 12 * 228:  # eud possible
        unit = player // 12
        player = player % 12
        return Deaths(player, comparison, number, unit)

    else:  # use epd
        return Deaths(player, comparison, number, 0)


# actions
def SetDeaths(player, settype, number, unit):
    return Action(
        0x00000000, 0x00000000, 0x00000000, 0x00000000,
        player, number, unit, 0x2D, settype, 0x14)


def SetMemory(offset, settype, number):
    ut.ep_assert(offset % 4 == 0)
    player = EPD(offset)

    if 0 <= player < 12 * 228:  # eud possible
        unit = player // 12
        player = player % 12
        return SetDeaths(player, settype, number, unit)

    else:  # use epd
        return SetDeaths(player, settype, number, 0)


def DisplayTextMessage(Text):
    return Action(0, Text, 0, 0, 0, 0, 0, 9, 0, 4)


def Draw():
    return Action(0, 0, 0, 0, 0, 0, 0, 56, 0, 4)


def EPD(offset):
    ut.ep_assert(offset % 4 == 0)
    return (offset - 0x0058A364) // 4



from setuptools import setup, find_packages
try:
    from Cython.Build import cythonize
except ImportError:
    def cythonize(x):
        return None

__version__ = '0.55.2'


setup(
    name="eudplib",
    version=__version__,
    packages=find_packages(),
    package_data={
        '': ['*.c', '*.pyx', '*.dll', '*.lst'],
    },
    ext_modules=cythonize([
        "eudplib/core/allocator/*.pyx",
        "eudplib/utils/*.pyx",
    ]),

    # metadata for upload to PyPI
    author="Trgk",
    author_email="whyask37@naver.com",
    description="EUD Trigger generator",
    license="MIT license",
    keywords="starcraft rawtrigger eud",
    url="http://blog.naver.com/whyask37/",  # project home page, if any
)
